/*
 * @(#)dlfcn.c	1.2 revision of 93/02/26  09:38:22
 * This is an unpublished work copyright (c) 1993 HELIOS Software GmbH
 * 3000 Hannover 1, Germany
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <dl.h>
#include "dlfcn.h"

/*
 * The void * handle returned from dlopen is actually a ModulePtr.
 */
typedef struct Module {
	struct Module	*next;
	char		*name;		/* module name for refcounting */
	int		refCnt;		/* the number of references */
	shl_t		shlHandle;	/* library handle from shl_load */
	struct dl_info	*info;		/* optional init/terminate functions */
} Module, *ModulePtr;

/*
 * We keep a list of all loaded modules to be able to call the fini
 * handlers at atexit() time.
 */
static ModulePtr modList;

/*
 * The last error from one of the dl* routines is kept in static
 * variables here. Each error is returned only once to the caller.
 */
static char errbuf[BUFSIZ];
static int errvalid;

extern char *strdup(const char *);
static void terminate(void);

/* ARGSUSED */
void *dlopen(const char *path, int mode)
{
	register ModulePtr mp;
	static int beenHere = 0;

	/*
	 * Upon the first call register a terminate handler that will
	 * close all libraries.
	 */
	if (!beenHere) {
		beenHere = 1;
		atexit(terminate);
	}
	/*
	 * Scan the list of modules if have the module already loaded.
	 */
	for (mp = modList; mp; mp = mp->next)
		if (strcmp(mp->name, path) == 0) {
			mp->refCnt++;
			return mp;
		}
	if ((mp = (ModulePtr)calloc(1, sizeof(*mp))) == NULL) {
		errvalid++;
		strcpy(errbuf, "calloc: ");
		strcat(errbuf, strerror(errno));
		return NULL;
	}
	if ((mp->name = strdup(path)) == NULL) {
		errvalid++;
		strcpy(errbuf, "strdup: ");
		strcat(errbuf, strerror(errno));
		free(mp);
		return NULL;
	}
	if ((mp->shlHandle = shl_load((char *)path,
	    (mode == RTLD_NOW ? BIND_IMMEDIATE : BIND_DEFERRED)|BIND_FIRST,
	    NULL)) == NULL) {
		free(mp->name);
		free(mp);
		errvalid++;
		strcpy(errbuf, "dlopen: ");
		strcat(errbuf, path);
		strcat(errbuf, ": ");
		strcat(errbuf, strerror(errno));
		return NULL;
	}
	mp->refCnt = 1;
	mp->next = modList;
	modList = mp;
	/*
	 * If there is a dl_info structure, call the init function.
	 */
	if (mp->info = (struct dl_info *)dlsym(mp, "dl_info")) {
		if (mp->info->init)
			(*mp->info->init)();
	} else
		errvalid = 0;
	return mp;
}

void *dlsym(void *handle, const char *symbol)
{
	register ModulePtr mp = (ModulePtr)handle;
	shl_t tmp = mp->shlHandle;
	long val;

	if (shl_findsym(&tmp, (char *)symbol, TYPE_UNDEFINED, &val) == -1) {
		errvalid++;
		strcpy(errbuf, "dlsym: ");
		strcat(errbuf, symbol);
		strcat(errbuf, ": ");
		strcat(errbuf, strerror(errno));
		return NULL;
	}
	return (void *)val;
}

char *dlerror(void)
{
	if (errvalid) {
		errvalid = 0;
		return errbuf;
	}
	return NULL;
}

int dlclose(void *handle)
{
	register ModulePtr mp = (ModulePtr)handle;
	int result;
	register ModulePtr mp1;

	if (--mp->refCnt > 0)
		return 0;
	if (mp->info && mp->info->fini)
		(*mp->info->fini)();
	result = shl_unload(mp->shlHandle);
	if (result == -1) {
		errvalid++;
		strcpy(errbuf, strerror(errno));
	}
	if (mp == modList)
		modList = mp->next;
	else {
		for (mp1 = modList; mp1; mp1 = mp1->next)
			if (mp1->next == mp) {
				mp1->next = mp->next;
				break;
			}
	}
	free(mp->name);
	free(mp);
	return result;
}

static void terminate(void)
{
	while (modList)
		dlclose(modList);
}
